from __future__ import annotations

import asyncio
from pathlib import Path
from typing import Optional

from async_bing_client import (
    Bing_Client,
    ConversationStyle,
    Text,
    SuggestRely,
    SourceAttribution,
    Apology,
    Image,
    Limit,
)
from ..nonebot_plugin_saa import Image as SaaImage
from ..nonebot_plugin_saa import MessageFactory
from ..nonebot_plugin_saa import Text as SaaText
from ..nonebot_plugin_spark_gpt.type_store.msgs_link import msg_links

from . import register_chatbot
from .BaseChatBot import BaseChatBot, Permission
from ..type_store.web_config import bing_config, common_config
from ..utils import get_url, md_to_pic

Client: Bing_Client = None
Cookie_Hash = hash(str(bing_config().cookie))
LOCKS = {}


async def ensure_client():
    global Client, Cookie_Hash
    config = bing_config()

    if not bool(config.cookie):
        raise Exception("模型Bing 已被禁用,目前无法使用")

    cookie_hash = hash(str(config.cookie))
    if (not Client) or cookie_hash != Cookie_Hash:
        Client = await Bing_Client(cookie=config.cookie, proxy=config.proxy).init()
        Cookie_Hash = cookie_hash


@register_chatbot
class BingBot(BaseChatBot):
    model = "Bing"
    permission = Permission.common
    chat: Optional[dict] = {}
    suggestions: Optional[list[str]] = []

    async def process(self, question):
        # 确保client最新
        await ensure_client()

        # 确保有chat
        if not self.chat:
            await self.refresh()

        # 确保有锁
        if self.__hash__() not in LOCKS.keys():
            LOCKS[self.__hash__()] = asyncio.Lock()

        # 判断是不是建议回复
        if question in ["1", "2", "3"]:
            if len(self.suggestions) >= int(question):
                question = self.suggestions[int(question) - 1]

        if self.command:
            question = self.command + "\n" + question

        return question

    @classmethod
    @property
    def desc(self) -> str:
        return "微软家Bing的gpt4联网模型"

    @classmethod
    @property
    def able(cls):
        return bool(bing_config().cookie)

    async def refresh(self):
        await ensure_client()
        if self.__hash__() not in LOCKS.keys():
            LOCKS[self.__hash__()] = asyncio.Lock()

        async with LOCKS[self.__hash__()]:
            from ..type_store import users

            try:
                self.chat = await Client.create_chat()
                users[self.owner_id].save()
            except Exception as e:
                raise Exception(f"Bing在创建新会话窗口时出错:{e}")

    async def ask_plain(
        self,
        question: str,
        image: str | Path | bytes = None,
    ):
        config = bing_config()
        question = await self.process(question)
        async with LOCKS[self.__hash__()]:
            full_answer = ""
            suggest_text = "\n\n建议回复:  \n"
            suggest_index = 0
            source_text = "\n\n资源链接:  \n"
            source_index = 0
            limit_text = ""
            limit: bool = False
            images = []
            async for data in Client.ask_stream_raw(
                question=question,
                personality=self.prompt,
                conversation_style=ConversationStyle.Creative,
                image=image,
                chat=self.chat,
            ):
                if isinstance(data, Text):
                    full_answer += str(data)
                elif isinstance(data, SuggestRely) and config.suggest_able:
                    self.suggestions.append(str(data))
                    suggest_index += 1
                    suggest_text += f"{suggest_index}.{data}  \n"
                elif isinstance(data, SourceAttribution):
                    source_index += 1
                    source_text += (
                        f"{source_index}.{data.display_name}:{data.see_more_url}\n"
                    )
                elif isinstance(data, Apology):
                    receipt = await MessageFactory(SaaText("\n" + data.content)).send(
                        at_sender=True
                    )
                    msg_links.add_msglink(self.owner_id, receipt.message_id, self)

                elif isinstance(data, Image):
                    images.append(SaaImage(data.url, name=f"{len(images) + 1}.png"))
                elif isinstance(data, Limit):
                    limit_text = f"\n连续对话上限:{data.num_user_messages}/{data.max_num_user_messages}  "
                    if data.num_user_messages >= data.max_num_user_messages:
                        limit = True

                if len(images) >= 4:
                    receipt = await MessageFactory(images).send(at_sender=True)
                    msg_links.add_msglink(self.owner_id, receipt.message_id, self)
                    images = []

            if images:
                receipt = await MessageFactory(images).send(at_sender=True)
                msg_links.add_msglink(self.owner_id, receipt.message_id, self)
            if len(source_text) > 10:
                full_answer += source_text
            if len(suggest_text) > 10 and config.suggest_able:
                full_answer += suggest_text
            if limit_text:
                full_answer += limit_text
            message = []

            # 并行来文转图并获取url
            async def md_to_pic_task(s, width=common_config().pic_width):
                if (self.auto_pic and len(full_answer) > self.num_limit) or self.pic:
                    pic = await md_to_pic(s, width)
                    return SaaImage(pic)
                else:
                    return SaaText(s)

            async def get_url_task(s):
                if self.url:
                    url = await get_url(s)
                    return SaaText(url)

            tasks = []
            tasks.append(asyncio.create_task(md_to_pic_task(full_answer)))
            tasks.append(asyncio.create_task(get_url_task(full_answer)))

            results = await asyncio.gather(*tasks)

            for result in results:
                if result:
                    message.append(result)

            receipt = await MessageFactory(message).send(at_sender=True)
            msg_links.add_msglink(self.owner_id, receipt.message_id, self)
        if limit:
            await self.refresh()
            receipt = await MessageFactory("聊天次数达到上限,已自动刷新对话").send(at_sender=True)
            msg_links.add_msglink(self.owner_id, receipt.message_id, self)

    async def ask_stream(self, question: str, image: str | Path | bytes = None):
        question = await self.process(question)

        async with LOCKS[self.__hash__()]:
            receipt = None
            answer = ""
            full_answer = ""
            suggest_text = "\n\n建议回复:  \n"
            suggest_index = 0
            source_text = "\n\n资源链接:  \n"
            source_index = 0
            limit_text = ""
            images = []
            limit: bool = False
            times = 0
            async for data in Client.ask_stream_raw(
                question=question,
                personality=self.prompt,
                conversation_style=ConversationStyle.Creative,
                image=image,
                chat=self.chat,
            ):
                if isinstance(data, Text):
                    answer += str(data)
                    full_answer += str(data)
                elif isinstance(data, SuggestRely):
                    self.suggestions.append(str(data))
                    suggest_index += 1
                    suggest_text += f"{suggest_index}. {data}  \n"
                elif isinstance(data, SourceAttribution):
                    source_index += 1
                    source_text += (
                        f"{source_index}.{data.display_name}:{data.see_more_url}\n"
                    )
                elif isinstance(data, Apology):
                    answer += str(data)
                    full_answer += str(data)
                elif isinstance(data, Image):
                    images.append(SaaImage(data.url, name=f"{len(images) + 1}.png"))
                elif isinstance(data, Limit):
                    limit_text = f"\n连续对话上限:{data.num_user_messages}/{data.max_num_user_messages}  "
                    if data.num_user_messages >= data.max_num_user_messages:
                        limit = True
                if "\n" in answer:
                    answer_slice = answer.rsplit("\n", 1)
                    if len(answer_slice[0].replace(" ", "").replace("\n", "")) > 80:
                        if receipt and receipt.edit_able:
                            await receipt.edit(SaaText(full_answer), at_sender=True)
                        else:
                            times += 1
                            receipt = await MessageFactory(
                                SaaText(answer_slice[0])
                            ).send(at_sender=True)
                            msg_links.add_msglink(
                                self.owner_id, receipt.message_id, self
                            )

                        answer = answer_slice[1]

                if len(images) >= 4:
                    receipt = await MessageFactory(images).send(at_sender=True)
                    msg_links.add_msglink(self.owner_id, receipt.message_id, self)
                    images = []

            next_text = ""

            if answer:
                next_text += answer
                full_answer += answer
                if receipt and receipt.edit_able:
                    await receipt.edit(SaaText(full_answer), at_sender=True)
            if images:
                receipt = await MessageFactory(images).send(at_sender=True)
                msg_links.add_msglink(self.owner_id, receipt.message_id, self)

            if len(source_text) > 10:
                full_answer += source_text
                next_text += source_text
            if len(suggest_text) > 10:
                full_answer += suggest_text
                next_text += suggest_text
            if limit_text:
                full_answer += limit_text
                next_text += limit_text
            if receipt and receipt.edit_able:
                await receipt.edit(SaaText(full_answer), at_sender=True)
            else:
                times += 1
                receipt = await MessageFactory(SaaText(next_text)).send(at_sender=True)
                msg_links.add_msglink(self.owner_id, receipt.message_id, self)

            if times > 1 and self.url:
                url = await get_url(full_answer)
                if receipt and receipt.edit_able:
                    await receipt.edit(SaaText(full_answer), at_sender=True)
                else:
                    receipt = await MessageFactory(SaaText(url)).send(at_sender=True)
                    msg_links.add_msglink(self.owner_id, receipt.message_id, self)

        if limit:
            await self.refresh()
            receipt = await MessageFactory("聊天次数达到上限,已自动刷新对话").send(at_sender=True)
            msg_links.add_msglink(self.owner_id, receipt.message_id, self)



from functools import partial
from io import BytesIO
from pathlib import Path
from typing import List

from nonebot.adapters import Bot as BaseBot
from nonebot.adapters import Event

from ..types import Text, Image, Reply, Mention
from ..utils import (
    Receipt,
    TargetQQGroup,
    MessageFactory,
    PlatformTarget,
    TargetQQPrivate,
    TargetOB12Unknow,
    QQGuildDMSManager,
    SupportedAdapters,
    SupportedPlatform,
    TargetQQGuildDirect,
    TargetQQGuildChannel,
    MessageSegmentFactory,
    get_bot_id,
    register_sender,
    register_get_bot_id,
    register_ms_adapter,
    register_qqguild_dms,
    register_list_targets,
    register_convert_to_arg,
    assamble_message_factory,
    register_target_extractor,
)

try:
    from nonebot.adapters.onebot.v12.exception import UnsupportedAction
    from nonebot.adapters.onebot.v12 import (
        Bot,
        Message,
        MessageEvent,
        MessageSegment,
        GroupMessageEvent,
        ChannelCreateEvent,
        ChannelDeleteEvent,
        ChannelMessageEvent,
        FriendDecreaseEvent,
        FriendIncreaseEvent,
        PrivateMessageEvent,
        GroupMessageDeleteEvent,
        GroupMemberDecreaseEvent,
        GroupMemberIncreaseEvent,
        ChannelMessageDeleteEvent,
        PrivateMessageDeleteEvent,
        ChannelMemberDecreaseEvent,
        ChannelMemberIncreaseEvent,
    )

    adapter = SupportedAdapters.onebot_v12
    register_onebot_v12 = partial(register_ms_adapter, adapter)

    MessageFactory.register_adapter_message(adapter, Message)


    @register_onebot_v12(Text)
    def _text(t: Text) -> MessageSegment:
        return MessageSegment.text(t.data["text"])


    @register_onebot_v12(Image)
    async def _image(i: Image, bot: BaseBot) -> MessageSegment:
        if not isinstance(bot, Bot):
            raise TypeError(f"Unsupported type of bot: {type(bot)}")
        image = i.data["image"]
        name = i.data["name"]
        if isinstance(image, str):
            resp = await bot.upload_file(type="url", name=name, url=image)
        elif isinstance(image, Path):
            resp = await bot.upload_file(
                type="path", name=name, path=str(image.resolve())
            )
        elif isinstance(image, BytesIO):
            image = image.getvalue()
            resp = await bot.upload_file(type="data", name=name, data=image)
        elif isinstance(image, bytes):
            resp = await bot.upload_file(type="data", name=name, data=image)
        else:
            raise TypeError(f"Unsupported type of image: {type(image)}")

        file_id = resp["file_id"]
        return MessageSegment.image(file_id)


    @register_onebot_v12(Mention)
    async def _mention(m: Mention) -> MessageSegment:
        return MessageSegment.mention(m.data["user_id"])


    @register_onebot_v12(Reply)
    async def _reply(r: Reply) -> MessageSegment:
        return MessageSegment.reply(r.data["message_id"])


    @register_target_extractor(PrivateMessageEvent)
    def _extract_private_msg_event(event: Event) -> PlatformTarget:
        assert isinstance(event, PrivateMessageEvent)
        if event.self.platform == "qq":
            return TargetQQPrivate(user_id=int(event.user_id))
        if event.self.platform == "qqguild":
            event_dict = event.dict()
            return TargetQQGuildDirect(
                recipient_id=int(event.user_id),
                source_guild_id=event_dict["qqguild"]["src_guild_id"],
            )
        return TargetOB12Unknow(
            platform=event.self.platform, detail_type="private", user_id=event.user_id
        )


    @register_target_extractor(GroupMessageEvent)
    def _extract_group_msg_event(event: Event) -> PlatformTarget:
        assert isinstance(event, GroupMessageEvent)
        if event.self.platform == "qq":
            return TargetQQGroup(group_id=int(event.group_id))
        return TargetOB12Unknow(
            platform=event.self.platform, detail_type="group", group_id=event.group_id
        )


    @register_target_extractor(ChannelMessageEvent)
    def _extarct_channel_msg_event(event: Event) -> PlatformTarget:
        assert isinstance(event, ChannelMessageEvent)
        if event.self.platform == "qqguild":  # all4one
            return TargetQQGuildChannel(channel_id=int(event.channel_id))
        return TargetOB12Unknow(
            platform=event.self.platform,
            detail_type="channel",
            channel_id=event.channel_id,
            guild_id=event.guild_id,
        )


    @register_target_extractor(FriendIncreaseEvent)
    @register_target_extractor(FriendDecreaseEvent)
    @register_target_extractor(PrivateMessageDeleteEvent)
    def _extract_private_notice_event(event: Event) -> PlatformTarget:
        assert isinstance(
            event, (FriendIncreaseEvent, FriendDecreaseEvent, PrivateMessageDeleteEvent)
        )
        if event.self.platform == "qq":
            return TargetQQPrivate(user_id=int(event.user_id))
        return TargetOB12Unknow(
            platform=event.self.platform, detail_type="private", user_id=event.user_id
        )


    @register_target_extractor(GroupMemberIncreaseEvent)
    @register_target_extractor(GroupMemberDecreaseEvent)
    @register_target_extractor(GroupMessageDeleteEvent)
    def _extract_group_notice_event(event: Event) -> PlatformTarget:
        assert isinstance(
            event,
            (
                GroupMemberIncreaseEvent,
                GroupMemberDecreaseEvent,
                GroupMessageDeleteEvent,
            ),
        )
        if event.self.platform == "qq":
            return TargetQQGroup(group_id=int(event.group_id))
        return TargetOB12Unknow(
            platform=event.self.platform, detail_type="group", group_id=event.group_id
        )


    @register_target_extractor(ChannelMemberIncreaseEvent)
    @register_target_extractor(ChannelMemberDecreaseEvent)
    @register_target_extractor(ChannelMessageDeleteEvent)
    @register_target_extractor(ChannelCreateEvent)
    @register_target_extractor(ChannelDeleteEvent)
    def _extarct_channel_notice_event(event: Event) -> PlatformTarget:
        assert isinstance(
            event,
            (
                ChannelMemberIncreaseEvent,
                ChannelMemberDecreaseEvent,
                ChannelMessageDeleteEvent,
                ChannelCreateEvent,
                ChannelDeleteEvent,
            ),
        )
        if event.self.platform == "qqguild":  # all4one
            return TargetQQGuildChannel(channel_id=int(event.channel_id))
        return TargetOB12Unknow(
            platform=event.self.platform,
            detail_type="channel",
            channel_id=event.channel_id,
            guild_id=event.guild_id,
        )


    @register_convert_to_arg(adapter, SupportedPlatform.qq_group)
    def _to_qq_group(target: PlatformTarget):
        assert isinstance(target, TargetQQGroup)
        return {
            "detail_type": "group",
            "group_id": str(target.group_id),
        }


    @register_convert_to_arg(adapter, SupportedPlatform.qq_private)
    def _to_qq_private(target: PlatformTarget):
        assert isinstance(target, TargetQQPrivate)
        return {
            "detail_type": "private",
            "user_id": str(target.user_id),
        }


    @register_convert_to_arg(adapter, SupportedPlatform.qq_guild_channel)
    def _to_qq_guild_channel(target: PlatformTarget):
        assert isinstance(target, TargetQQGuildChannel)
        return {
            "detail_type": "channel",
            "channel_id": str(target.channel_id),
        }


    @register_convert_to_arg(adapter, SupportedPlatform.qq_guild_direct)
    def _to_qq_guild_direct(target: PlatformTarget):
        assert isinstance(target, TargetQQGuildDirect)
        return {
            "detail_type": "private",
            "guild_id": str(QQGuildDMSManager.get_guild_id(target)),
        }


    @register_qqguild_dms(adapter)
    async def _qqguild_dms(target: TargetQQGuildDirect, bot: BaseBot) -> int:
        assert isinstance(bot, Bot)

        resp = await bot.create_dms(
            user_id=str(target.recipient_id), src_guild_id=str(target.source_guild_id)
        )
        return resp["guild_id"]


    @register_convert_to_arg(adapter, SupportedPlatform.unknown_ob12)
    def _to_unknow(target: PlatformTarget):
        assert isinstance(target, TargetOB12Unknow)
        return target.dict(exclude={"platform", "platform_type"})


    class OB12Receipt(Receipt):
        message_id: str
        adapter_name = adapter


        async def revoke(self):
            return await self._get_bot().delete_message(message_id=self.message_id)

        @property
        def raw(self):
            return self.message_id



    @register_sender(SupportedAdapters.onebot_v12)
    async def send(
        bot,
        msg: MessageFactory[MessageSegmentFactory],
        target,
        event,
        at_sender: bool,
        reply: bool,
    ) -> OB12Receipt:
        assert isinstance(bot, Bot)
        assert isinstance(
            target,
            (
                TargetQQGroup,
                TargetQQPrivate,
                TargetQQGuildChannel,
                TargetQQGuildDirect,
                TargetOB12Unknow,
            ),
        )

        if event:
            assert isinstance(event, MessageEvent)
            full_msg = assamble_message_factory(
                msg, Mention(event.user_id), Reply(event.message_id), at_sender, reply
            )
        else:
            full_msg = msg
        msg_to_send = await full_msg.build(bot)
        assert isinstance(msg_to_send, Message)
        if bot.platform == "qqguild":
            assert isinstance(target, (TargetQQGuildChannel, TargetQQGuildDirect))
            if isinstance(target, TargetQQGuildDirect):
                await QQGuildDMSManager.aget_guild_id(target, bot)
            params = {}
            if event:
                # 传递 event_id，用来支持频道的被动消息
                params["event_id"] = event.id
            resp = await bot.send_message(
                message=msg_to_send,
                **target.arg_dict(bot),
                **params,
            )
        else:
            resp = await bot.send_message(message=msg_to_send, **target.arg_dict(bot))
        message_id = resp["message_id"]
        return OB12Receipt(bot_id=get_bot_id(bot), message_id=message_id)


    @register_list_targets(SupportedAdapters.onebot_v12)
    async def list_targets(bot: BaseBot) -> List[PlatformTarget]:
        assert isinstance(bot, Bot)

        targets = []
        try:
            friends = await bot.get_friend_list()
            for friend in friends:
                platform = bot.platform
                if platform == "qq":
                    targets.append(TargetQQPrivate(user_id=int(friend["user_id"])))
                elif platform == "qqguild":
                    # FIXME: 怎么获取 src_guild_id 捏？
                    pass
                else:
                    targets.append(
                        TargetOB12Unknow(
                            platform=platform,
                            detail_type="private",
                            user_id=friend["user_id"],
                        )
                    )

        except UnsupportedAction:  # pragma: no cover
            pass

        try:
            groups = await bot.get_group_list()
            for group in groups:
                platform = bot.platform
                if platform == "qq":
                    targets.append(TargetQQGroup(group_id=int(group["group_id"])))
                else:
                    targets.append(
                        TargetOB12Unknow(
                            platform=platform,
                            detail_type="group",
                            group_id=group["group_id"],
                        )
                    )

        except UnsupportedAction:  # pragma: no cover
            pass

        try:
            guilds = await bot.get_guild_list()
            for guild in guilds:
                channels = await bot.get_channel_list(guild_id=guild["guild_id"])
                for channel in channels:
                    platform = bot.platform
                    if platform == "qqguild":
                        targets.append(
                            TargetQQGuildChannel(channel_id=int(channel["channel_id"]))
                        )
                    else:
                        targets.append(
                            TargetOB12Unknow(
                                platform=platform,
                                detail_type="channel",
                                channel_id=channel["channel_id"],
                                guild_id=guild["guild_id"],
                            )
                        )

        except UnsupportedAction:  # pragma: no cover
            pass

        return targets



    @register_get_bot_id(adapter)
    def _get_bot_id(bot: BaseBot):
        assert isinstance(bot, Bot)
        return f"{bot.platform}-{bot.self_id}"

except ImportError:
    pass
except Exception as e:
    raise e



import json
from abc import ABC
from typing import Any, Dict, Type, ClassVar, cast, Union, Optional

from nonebot.adapters import Bot
from pydantic import BaseModel, ConfigDict

from .auto_select_bot import get_bot_by_id
from .const import SupportedAdapters
from .exceptions import AdapterNotSupported



class Receipt(BaseModel, ABC):
    model_config = ConfigDict(frozen=True, orm_mode=True)

    _deserializer_map: ClassVar[Dict[SupportedAdapters, Type["Receipt"]]] = {}

    def __init_subclass__(cls) -> None:
        assert isinstance(cls.__fields__["adapter_name"].default, SupportedAdapters)
        cls._deserializer_map[cls.__fields__["adapter_name"].default] = cls
        return super().__init_subclass__()

    @classmethod
    def deserialize(cls, source: Any) -> "Receipt":
        if isinstance(source, str):
            raw_obj = json.loads(source)
        else:
            raw_obj = source
            assert raw_obj.get("adapter_name")
        adapter_name = cast(
            SupportedAdapters, SupportedAdapters(raw_obj["adapter_name"])
        )
        return cls._deserializer_map[adapter_name].parse_obj(raw_obj)

    message_id: Union[str, int]
    sent_msg: Optional[Any]

    adapter_name: SupportedAdapters
    bot_id: str

    def _get_bot(self) -> Bot:
        bot = get_bot_by_id(self.adapter_name, self.bot_id)
        if not bot:
            raise RuntimeError(
                f"Bot of {self.adapter_name} {self.bot_id} not connected"
            )
        return bot

    async def revoke(self):
        ...

    async def edit(self, msg, at_sender=False, reply=False):
        raise AdapterNotSupported(self._get_bot().adapter.get_name())

    @classmethod
    @property
    def revoke_able(cls):
        return getattr(Receipt, 'revoke') != getattr(cls, 'revoke')

    @classmethod
    @property
    def edit_able(cls):
        return getattr(Receipt, 'edit') != getattr(cls, 'edit')


    @property
    def raw(self) -> Any:
        ...



from nonebot_plugin_templates.template_types import Menus, Menu, Funcs, Func
from nonebot_plugin_templates.templates_render import menu_render

from ..type_store.web_config import common_config
from ..utils import Command_Start


class SparkMenus:
    images: dict = {}
    chat: Menus = Menus(
        menus=Menu(
            "会话",
            des="会话的使用和管理\n下面的所有命令中,带{}号的为需要手动填写的值,括号中的参数为选填.",
            funcs=Funcs(
                Func(name=common_config().public_command + "{chat_name} {question} \n( {photo} )",
                     desc="使用公有会话询问\n-参数:\n--必填:\n---1.chat_name: 会话昵称\n---2.question: 问题\n--选填:\n---1. photo:需要识别的图片内容,只有model为bing时才可以使用,需要跟随文本信息一同发送")
                + Func(name=common_config().private_command + "{chat_name} {question} \n( {photo} )",
                       desc="使用私有会话询问\n-参数:\n--必填:\n---1.chat_name: 会话昵称\n---2.question: 问题\n--选填:\n---1. photo:需要识别的图片内容,只有model为bing时才可以使用,需要跟随文本信息一同发送")
                + Func(
                    name=Command_Start + "chat add -n {chat_name} -m {model}   \n( -p {prompt} -c {command} -auto_pic {bool} -num_limit {int} -pic {bool} -url {bool} -stream {bool} -public )",
                    desc="创建一个新的会话:\n-参数:\n--必填:\n---1.chat_name: 会话名称\n---2.model: 使用的模型名称或者索引数字 \n--选填:\n---1.prompt:\n----使用本地预设,须在预设名称前加'.'号\n----使用自定义预设,直接输入内容\n---2.command: \n----使用本地指令,须在指令名称前加'.'号\n----使用自定义指令,直接输入内容\n---3.auto_pic:是否当字数过多时自动转图片\n---4.num_limit:自动转图片的字数上限\n---5.pic:是否总是文字转图片\n---6.url:是否在图片或流式回复时发送全文链接\n---7.stream:是否流式输出,分段发送结果",
                )
                + Func(
                    name=Command_Start + "chat edit -on {old_name}   \n( -nn {new_name} -m {model} -p {prompt} -c {command} -auto_pic {bool} -num_limit {int} -pic {bool} -url {bool} -stream {bool} -public )",
                    desc="修改一个会话(只会修改选择的参数):\n-参数:\n--必填:\n---1.old_name: 原来的会话名称\n--选填:\n---1.new_name: 新的会话名称\n---2.model: 使用的模型名称或者索引数字 \n---3.prompt: 预设\n---4.command: 指令\n---5.auto_pic:是否当字数过多时自动转图片\n---6.num_limit:自动转图片的字数上限\n---7.pic:是否总是文字转图片\n---8.url:是否在图片或流式回复时发送全文链接\n---9.stream:是否流式输出,分段发送结果",
                )
                + Func(name=Command_Start + "chat list", desc="以图片形式查看会话列表,包括私有列表和公有列表")
                + Func(name=Command_Start + "chat del -n {chat_name}   \n( -public )",
                       desc="删除一个会话:\n-参数:\n--chat_name: 会话名称")
            ),
        )
    )

    model: Menus = Menus(
        menus=Menu(
            "模型",
            des="语言模型: 指会话生成回复所使用的模型,不同模型有不同的特性",
            funcs=Funcs(Func(name=Command_Start + "model list", desc="以图片形式展示所有的可使用的模型及其介绍")),
        )
    )

    prompt: Menus = Menus(
        menus=Menu(
            "预设",
            des="预设: 指每次新对话开始前,优先加载的人格\n下面的所有命令中,带{}号的为需要手动填写的值,括号中的参数为选填.",
            funcs=Funcs(
                Func(
                    name=Command_Start + "prompt add -n {name} -c {content}",
                    desc="添加一个新的预设\n-参数:\n--1.name: 预设名称\n--2.content: 预设内容",
                )
                + Func(
                    name=Command_Start + "prompt del -n {name}",
                    desc="删除一个预设\n-参数:\n--name: 预设名称",
                )
                + Func(
                    name=Command_Start + "prompt list   \n( -p {num} )",
                    desc="以图片形式查看预设列表\n-参数:\n--num: 页数(默认第一页)\n",
                )
                + Func(
                    name=Command_Start + "prompt show -n {name}",
                    desc="以图片形式展示一个预设的完整内容\n-参数:\n--name:预设名称",
                )
            ),
        )
    )

    command: Menus = Menus(
        menus=Menu(
            "指令",
            des="指令: 指自动填充在问题前面的指令,可以用来指定回复的格式或者要求\n下面的所有命令中,带{}号的为需要手动填写的值,括号中的参数为选填.",
            funcs=Funcs(
                Func(
                    name=Command_Start + "command add -n {name} -c {content}",
                    desc="添加一个新的指令\n-参数:\n--1.name: 指令名称\n--2.content: 指令内容",
                )
                + Func(
                    name=Command_Start + "command del -n {name}",
                    desc="删除一个指令\n-参数:\n--name: 指令名称",
                )
                + Func(
                    name=Command_Start + "command list   \n( -p {num} )",
                    desc="以图片形式查看指令列表\n-参数:\n--num: 页数",
                )
                + Func(
                    name=Command_Start + "command show -n {name}",
                    desc="以图片形式展示一个指令的完整内容\n-参数:\n--name:指令名称",
                )
            ),
        )
    )

    @property
    def all(self) -> Menus:
        return self.chat + self.model + self.prompt + self.command

    async def get_menu_image(self, menu_name) -> bytes:
        menus_: Menus = getattr(self, menu_name)
        if menus_:
            if menu_name in self.images.keys():
                return self.images[menu_name]
            else:
                image = await menu_render(menus_, width=850)
                self.images[menu_name] = image
                return image
        else:
            raise Exception("Menu not exist.")

    def get_menu_string(self, menu_name) -> str:
        menus_: Menus = getattr(self, menu_name)
        if menus_:
            result = ""
            for menu_index, menu in enumerate(menus_.menus):
                result += f"[{menu_index + 1}].{menu.name}功能\n详细说明请发送 '{menu_name} help' 查看图片版\n"  # noqa: E501
                for func_index, func in enumerate(menu.funcs):
                    result += f"  ({func_index + 1}): {func.name}\n"
                result += "\n"
            return result
        else:
            raise Exception("Menu not exist.")

    def get_func_string(self, func_name: str):
        if func_name == "private chat":
            return "使用私有会话询问\n-参数:\n--必填:\n---1.chat_name: 会话昵称\n---2.question: 问题\n--选填:\n---1. photo:需要识别的图片内容,只有model为bing时才可以使用,需要跟随文本信息一同发送"

        elif func_name == "public chat":
            return "使用公有会话询问\n-参数:\n--必填:\n---1.chat_name: 会话昵称\n---2.question: 问题\n--选填:\n---1. photo:需要识别的图片内容,只有model为bing时才可以使用,需要跟随文本信息一同发送"

        menu_name, func_name = func_name.split(".")
        menu = getattr(self, menu_name)
        for each_func in menu.menus[0].funcs.funcs:
            if each_func.name.startswith(Command_Start + "{menu_name} {func_name}"):
                return f"格式:{each_func.name}\n\n描述:\n{each_func.desc}"


menus = SparkMenus()
